﻿using System;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.ServiceProcess;
using Inedo.Consolation;
using Inedo.WindowsServices;

namespace Inedo.Agents.Service
{
    public static class Program
    {
        [ConsoleCommand]
        [Description("Runs the agent service interactively.")]
        public static void run([ConsoleAlias("s"), Browsable(false)] bool runAsService = false)
        {
            if (runAsService)
            {
                ServiceBase.Run(AgentService.Instance.GetWindowsService());
            }
            else
            {
                ConsoleCancelEventHandler cancel =
                    (s, e) =>
                    {
                        e.Cancel = true;
                        AgentService.Instance.Stop();
                    };

                AgentService.Instance.Start();
                Console.CancelKeyPress += cancel;

                try
                {
                    AgentService.Instance.Task.Wait();
                }
                finally
                {
                    Console.CancelKeyPress -= cancel;
                }
            }
        }

        [ConsoleCommand]
        [Description("Installs the agent service as a Windows service.")]
        public static void install([Description("The user name of the service account.")] string user = null, [Description("The password of the service account.")] string password = null)
        {
            using (var service = WindowsService.GetService(AgentService.Instance.Name))
            {
                if (service != null)
                {
                    Console.WriteLine($"Service {AgentService.Instance.Name} already exists; attempting to uninstall...");
                    service.Delete();
                    Console.WriteLine("Service removed.");
                }
            }

            var commandLine = $"\"{typeof(Program).Assembly.Location}\" run -s";

            Console.WriteLine($"Attempting to create service {AgentService.Instance.Name}...");
            using (var service = WindowsService.CreateService(AgentService.Instance.Name, AgentService.Instance.FriendlyName, commandLine, user, password))
            {
                service.Description = AgentService.Instance.Description;
                service.FailureActions = new ServiceFailureActions(null, Enumerable.Repeat(new ServiceControllerAction(ServiceControllerActionType.Restart, new TimeSpan(0, 1, 0)), 3).ToArray());
                Console.WriteLine("Service installed.");
            }
        }

        [ConsoleCommand]
        [Description("Uninstalls the agent Windows service.")]
        public static void uninstall()
        {
            using (var service = WindowsService.GetService(AgentService.Instance.Name))
            {
                if (service != null)
                {
                    try
                    {
                        using (var serviceProcess = new ServiceController(service.Name))
                        {
                            if (serviceProcess.CanStop)
                            {
                                Console.WriteLine($"Stopping service {service.Name}...");
                                serviceProcess.Stop();
                                serviceProcess.WaitForStatus(ServiceControllerStatus.Stopped);
                                Console.WriteLine($"Service stopped.");
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Could not stop service: " + ex.Message);
                    }

                    Console.WriteLine($"Uninstalling service {AgentService.Instance.Name}...");
                    service.Delete();
                    Console.WriteLine("Service removed.");
                }
                else
                {
                    Console.WriteLine($"Service {AgentService.Instance.Name} is not installed.");
                }
            }
        }

        [ConsoleCommand]
        [Description("Generates a random AES encryption key.")]
        public static void generateKey()
        {
            using (var random = new RNGCryptoServiceProvider())
            {
                var data = new byte[16];
                random.GetBytes(data);
                Console.WriteLine(AgentEndpoint.FormatEncryptionKey(data));
            }
        }

        [ConsoleCommand]
        [Browsable(false)]
        public static int update(string targetPath, string serviceName, int pid)
        {
            using (var updater = new ServerUpdater(Path.GetDirectoryName(typeof(Program).Assembly.Location), targetPath, serviceName, pid))
            {
                if (!updater.Update())
                {
                    updater.Rollback();
                    return -1;
                }
            }

            return 0;
        }

        public static int Main(string[] args) => ConsoleApplication.Run(typeof(Program), args);
    }
}
