﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;

namespace Inedo
{
    /// <summary>
    /// A collection of Additional Helper methods to perform reflection related tasks.
    /// </summary>
    public static class AhReflection
    {
        /// <summary>
        /// Returns the generic arguments from an ancestor with the specified generic type definition from the specified type, or an empty type array if the 
        /// type's ancestors do not implement the specified generic type definition.
        /// </summary>
        /// <param name="type">Type to inspect.</param>
        /// <param name="genericTypeDefinition">A generic type definition.</param>
        /// <returns>A type arguments array or an empty array.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="type"/> is null or <paramref name="genericTypeDefinition"/> is null.</exception>
        /// <exception cref="ArgumentException"><paramref name="genericTypeDefinition"/> is not a generic type definition.</exception>
        public static Type[] GetGenericArgumentsFromAncestor(Type type, Type genericTypeDefinition)
        {
            if (type == null)
                throw new ArgumentNullException("type");
            if (genericTypeDefinition == null)
                throw new ArgumentNullException("genericTypeDefinition");
            if (!genericTypeDefinition.IsGenericTypeDefinition)
                throw new ArgumentException("Not a generic type definition.", "genericTypeDefinition");

            if (genericTypeDefinition.IsInterface)
            {
                return type.GetInterfaces()
                    .Where(i => i.IsGenericType && i.GetGenericTypeDefinition() == genericTypeDefinition)
                    .Select(i => i.GetGenericArguments())
                    .FirstOrDefault() ?? Type.EmptyTypes;
            }
            else
            {
                do
                {
                    if (type.IsGenericType && type.GetGenericTypeDefinition() == genericTypeDefinition)
                        return type.GetGenericArguments();
                    type = type.BaseType;
                } while (type != null);
            }

            return Type.EmptyTypes;
        }

        /// <summary>
        /// Returns a value indicating whether a type implements an interface.
        /// </summary>
        /// <param name="toCheck">The type to check.</param>
        /// <param name="iface">The interface to search for.</param>
        /// <returns>True if the type implements the interface; otherwise false.</returns>
        /// <exception cref="System.ArgumentNullException">An argument is null.</exception>
        /// <exception cref="System.ArgumentException">The <paramref name="iface"/> parameter is not an interface type.</exception>
        public static bool ImplementsInterface(Type toCheck, Type iface)
        {
            if (toCheck == null)
                throw new ArgumentNullException("toCheck");
            if (iface == null)
                throw new ArgumentNullException("iface");
            if (!iface.IsInterface)
                throw new ArgumentException("The type is not an interface", "iface");

            if (!iface.IsGenericTypeDefinition)
                return iface.IsAssignableFrom(toCheck);

            foreach (var i in toCheck.GetInterfaces())
            {
                if (i.IsGenericType)
                {
                    if (i.GetGenericTypeDefinition() == iface)
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Returns an attribute (typed <typeparamref name="T"/>) if it is present on a member.
        /// </summary>
        /// <typeparam name="T">Type of attribute to return.</typeparam>
        /// <param name="member">Member to search for the attribute on.</param>
        /// <returns>Attribute if it was found; otherwise null.</returns>
        public static T GetCustomAttribute<T>(ICustomAttributeProvider member)
            where T : Attribute
        {
            if (member == null)
                throw new ArgumentNullException("member");

            return GetCustomAttributes<T>(member).FirstOrDefault();
        }

        /// <summary>
        /// Returns all instances of an attribute type if it is present on a member.
        /// </summary>
        /// <typeparam name="T">Type fo the attribute to return.</typeparam>
        /// <param name="member">Member to search for the attributes on.</param>
        /// <returns>Attributes that were found.</returns>
        public static IEnumerable<T> GetCustomAttributes<T>(ICustomAttributeProvider member)
            where T : Attribute
        {
            if (member == null)
                throw new ArgumentNullException("member");

            var attribs = member.GetCustomAttributes(typeof(T), true);
            return attribs.Cast<T>();
        }

        /// <summary>
        /// Returns an enumeration of the specified type and all its base types.
        /// </summary>
        /// <param name="type">The type to enumerate.</param>
        /// <returns>The specified type and all its base types.</returns>
        public static IEnumerable<Type> GetBaseTypes(Type type)
        {
            if (type == null)
                throw new ArgumentNullException("type");

            var currentType = type;
            while (currentType != null)
            {
                yield return currentType;
                currentType = currentType.BaseType;
            }
        }

        /// <summary>
        /// Returns the first matching method with a signature compatible to the specified argument types.
        /// </summary>
        /// <param name="type">The object type which contains the method.</param>
        /// <param name="methodName">The name of the method.</param>
        /// <param name="argTypes">The types of the arguments supplied to the method.</param>
        /// <returns>First matching <see cref="MethodInfo"/> or null if no matching method was found.</returns>
        public static MethodInfo GetMethodWithSignature(Type type, string methodName, IEnumerable<Type> argTypes)
        {
            if (type == null)
                throw new ArgumentNullException("type");
            if (string.IsNullOrEmpty(methodName))
                throw new ArgumentNullException("methodName");
            if (argTypes == null)
                throw new ArgumentNullException("argTypes");

            var paramTypes = argTypes.ToList();

            Func<ParameterInfo[], bool> areParamsCompatible =
                argParamTypes =>
                {
                    for (int i = 0; i < paramTypes.Count; i++)
                    {
                        if (paramTypes[i] != null)
                        {
                            if (!argParamTypes[i].ParameterType.IsAssignableFrom(paramTypes[i]))
                                return false;
                        }
                        else
                        {
                            if (!argParamTypes[i].ParameterType.IsClass && Nullable.GetUnderlyingType(argParamTypes[i].ParameterType) == null)
                                return false;
                        }
                    }

                    return true;
                };

            return type
                .GetMethods(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Static)
                .Where(m => m.Name == methodName)
                .Select(m => new { Method = m, Params = m.GetParameters() })
                .Where(m => m.Params.Length == paramTypes.Count && areParamsCompatible(m.Params))
                .Select(m => m.Method)
                .FirstOrDefault();
        }
    }
}
