﻿using System;

namespace Inedo.WindowsServices
{
    /// <summary>
    /// Represents an action that the service control manager can perform.
    /// </summary>
    [Undisclosed]
    public struct ServiceControllerAction : IEquatable<ServiceControllerAction>
    {
        private ServiceControllerActionType type;
        private int delay;

        /// <summary>
        /// Initializes a new instance of the <see cref="ServiceControllerAction"/> struct.
        /// </summary>
        /// <param name="type">The action to be performed.</param>
        /// <param name="delay">The time to wait before performing the specified action.</param>
        /// <exception cref="System.ArgumentOutOfRangeException"><paramref name="delay"/> is less than zero or greater than <see cref="Int32.MaxValue"/>.</exception>
        /// <remarks>
        /// The <paramref name="delay"/> parameter has a resolution of milliseconds.
        /// </remarks>
        public ServiceControllerAction(ServiceControllerActionType type, TimeSpan delay)
        {
            if (delay.TotalMilliseconds < 0 || delay.TotalMilliseconds > int.MaxValue)
                throw new ArgumentOutOfRangeException("delay");

            this.type = type;
            this.delay = (int)delay.TotalMilliseconds;
        }

        /// <summary>
        /// ==s the specified action1.
        /// </summary>
        /// <param name="action1">The action1.</param>
        /// <param name="action2">The action2.</param>
        /// <returns></returns>
        public static bool operator ==(ServiceControllerAction action1, ServiceControllerAction action2)
        {
            return action1.Equals(action2);
        }
        /// <summary>
        /// !=s the specified action1.
        /// </summary>
        /// <param name="action1">The action1.</param>
        /// <param name="action2">The action2.</param>
        /// <returns></returns>
        public static bool operator !=(ServiceControllerAction action1, ServiceControllerAction action2)
        {
            return !action1.Equals(action2);
        }

        /// <summary>
        /// Gets the action to be performed.
        /// </summary>
        public ServiceControllerActionType Type
        {
            get { return this.type; }
        }
        /// <summary>
        /// Gets the time to wait before performing the specified action.
        /// </summary>
        /// <remarks>
        /// This value has a resolution of milliseconds.
        /// </remarks>
        public TimeSpan Delay
        {
            get { return new TimeSpan(0, 0, 0, 0, this.delay); }
            set
            {
                if (value.TotalMilliseconds < 0 || value.TotalMilliseconds > int.MaxValue)
                    throw new ArgumentOutOfRangeException();

                this.delay = (int)value.TotalMilliseconds;
            }
        }

        /// <summary>
        /// Indicates whether the current object is equal to another object of the same type.
        /// </summary>
        /// <param name="other">An object to compare with this object.</param>
        /// <returns>
        /// true if the current object is equal to the <paramref name="other" /> parameter; otherwise, false.
        /// </returns>
        public bool Equals(ServiceControllerAction other)
        {
            return this.type == other.type && this.delay == other.delay;
        }
        /// <summary>
        /// Determines whether the specified <see cref="System.Object" /> is equal to this instance.
        /// </summary>
        /// <param name="obj">The <see cref="System.Object" /> to compare with this instance.</param>
        /// <returns>
        ///   <c>true</c> if the specified <see cref="System.Object" /> is equal to this instance; otherwise, <c>false</c>.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (obj is ServiceControllerAction)
                return this.Equals((ServiceControllerAction)obj);
            else
                return false;
        }
        /// <summary>
        /// Returns a hash code for this instance.
        /// </summary>
        /// <returns>
        /// A hash code for this instance, suitable for use in hashing algorithms and data structures like a hash table. 
        /// </returns>
        public override int GetHashCode()
        {
            return ((int)this.type << 24) ^ this.delay;
        }
        /// <summary>
        /// Returns a <see cref="System.String" /> that represents this instance.
        /// </summary>
        /// <returns>
        /// A <see cref="System.String" /> that represents this instance.
        /// </returns>
        public override string ToString()
        {
            string message;

            switch (this.Type)
            {
                case ServiceControllerActionType.None:
                    return "Do nothing";

                case ServiceControllerActionType.Reboot:
                    message = "Restart the computer";
                    break;

                case ServiceControllerActionType.Restart:
                    message = "Restart the service";
                    break;

                case ServiceControllerActionType.RunCommand:
                    message = "Run the specified command";
                    break;

                default:
                    return "Unknown";
            }

            return string.Format("{0} after {1}", message, this.Delay);
        }
    }
}
